/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: aushow.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sys/soundcard.h>
#include <sysexits.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "auwith_int.h"

#define RRR 4

static const char *
aulabels[] = SOUND_DEVICE_LABELS;

static const char *
aunames[] = SOUND_DEVICE_NAMES;

static const int
aumod[RRR] =
{ SOUND_MIXER_READ_DEVMASK
, SOUND_MIXER_READ_RECMASK
, SOUND_MIXER_READ_RECSRC
, SOUND_MIXER_READ_STEREODEVS
};

static const char *
aufnm[] =
{ "MU_LAW"   , "A_LAW"    , "IMA_ADPCM", "U8"
, "S16_LE"   , "S16_BE"   , "S8"       , "U16_LE"
, "U16_BE"   , "MPEG"     , "AC3"      , "unkn"
, "S32_LE"   , "S32_BE"   , "U32_LE"   , "U32_BE"

, "S24_LE"   , "S24_BE"   , "U24_LE"   , "U24_BE"
, "unkn"     , "unkn"     , "unkn"     , "unkn"
, "unkn"     , "unkn"     , "unkn"     , "unkn"
, "STEREO"   , "WEIRD"    , "unkn"     , "FULLDUPLEX"
};

static const char *
aucap[] =
{ "DUPLEX"   , "REALTIME", "BATCH"     , "COPROC"
, "TRIGGER"  , "MMAP"    , "MULTI"     , "BIND"

, "INPUT"    , "OUTPUT"  , "VIRTUAL"   , "unkn"
, "ANALOGOUT", "ANALOGIN", "DIGITALOUT", "DIGITALIN"
, "SPECIAL"  , "unkn"    , "unkn"      , "unkn"
, "unkn"     , "unkn"    , "unkn"      , "unkn"
};

static const char *
midicap[] =
{ "MPU401"  , "INPUT" , "OUTPUT", "VIRTUAL"
, "MTCINPUT", "CLIENT", "SERVER", "INTERNAL"
, "EXTERNAL", "PTOP"  , "MTC"   , "unkn"
, "unkn"    , "unkn"  , "unkn"  , "unkn"

, "unkn"    , "unkn"  , "unkn"  , "unkn"
, "unkn"    , "unkn"  , "unkn"  , "unkn"
, "unkn"    , "unkn"  , "unkn"  , "unkn"
, "unkn"    , "unkn"  , "unkn"  , "unkn"
};


static void
/*****************************************************************************************************/
incaps(FILE *fo, int cap) {                                                                      /****
 *****************************************************************************************************/
    u_int32_t i, m, k;

    fprintf(fo, "%08X=<", cap);
    for (k = 0, m = 1, i = 0; i < SOUND_MIXER_NRDEVICES; ++i, cap >>= 1) {
        if  (cap & m) fprintf(fo, "%s%s", k++ ? "," : "", aunames[i]);
    }
    if  (cap) fprintf(fo, "%s%s", k++ ? "," : "", "unkn");
    fprintf(fo, ">");
}

static void
/*****************************************************************************************************/
midicaps(FILE *fo, int cap) {                                                                    /****
 *****************************************************************************************************/
    u_int32_t i, m, k;

    fprintf(fo, "%08X=<", cap);
    for (k = 0, m = 1, i = 0; m; ++i, m <<= 1) {
        if  (cap & m) fprintf(fo, "%s%s", k++ ? "," : "", midicap[i]);
    }
    fprintf(fo, ">");
}

static void
/*****************************************************************************************************/
aucaps(FILE *fo, int cap) {                                                                      /****
 *****************************************************************************************************/
    u_int32_t i, m, k;

    fprintf(fo, "%08X=<rev=%u", cap, cap & 0xFF);
    for (k = 0, m = 0x100, i = 0; !!m; ++i, m <<= 1) {
        if  (cap & m) fprintf(fo, "%s%s", k++ ? "," : " ", aucap[i]);
    }
    fprintf(fo, ">");
}

static void
/*****************************************************************************************************/
aumixeinfo(FILE *fo, oss_mixer_enuminfo *mei) {                                                  /****
 *****************************************************************************************************/
    int i;

    fprintf( fo, "dev=%d ctrl=%d nvalues=%d version=%d <"
           , mei->dev, mei->ctrl, mei->nvalues, mei->version
           )
    ;
    for (i = 0; i < mei->nvalues; ++i) {
        fprintf(fo, "%s%s", i ? "," : "", &mei->strings[mei->strindex[i]]);
    }
    fprintf(fo, ">");
}

static void
/*****************************************************************************************************/
aufrm(FILE *fo, u_long frm) {                                                                    /****
 *****************************************************************************************************/
    u_int32_t i, m, k;

    fprintf(fo, "%08lX=<", frm);
    for (k = 0, m = 1, i = 0; m; ++i, m <<= 1) {
        if  (frm & m) fprintf(fo, "%s%s", k++ ? "," : "", aufnm[i]);
    }
    fprintf(fo, ">");
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_show(auwith_env *env, FILE *fo, u_int full) {                                             /****
 *****************************************************************************************************
 *****************************************************************************************************/
    char *mn[AUWITH_NFD] = {"mix", "dev"};
    struct snd_size blocksz;
    unsigned long long ull;
    oss_mixer_enuminfo mei;
    int c, k, ex = EX_OK;
    snd_capabilities sc;
    audio_buf_info bi;
    snd_chan_param sp;
    snd_sync_parm syp;
    buffmem_desc bmd;
    audio_errinfo ei;
    oss_count_t oct;
    count_info ci;

    fprintf(fo, "    |      DEVMASK, RECMASK,  RECSRC,  STEREO\n");
    for (c = 0; c < AUWITH_NFD; ++c) {
        if  ((1 << c) & full) {
            int i, m, r[RRR];

            for (i = 0; i < RRR; ++i) {
                r[i] = 0;
                if  (0 > ioctl(env->aufd[c], aumod[i], &r[i])) {
                    ifBLIN_QW0("audio mask %d %d", c, i);
                    EXOUT(EX_IOERR);
            }   }
            fprintf( fo
                   , "    | %s %08X %08X %08X %08X %s"
                   , mn[c], r[0], r[1], r[2], r[3], env->fnm[c]
                   );
            if  (c) {
                int bitps = 0, chan = 0, rate = 0, bsize = 0;

                if  (0 > ioctl(env->aufd[c], SOUND_PCM_READ_CHANNELS, &chan)) {
                    ifBLIN_QW0("audio chan %d", c);
                    EXOUT(EX_IOERR);
                }
                if  (0 > ioctl(env->aufd[c], SOUND_PCM_READ_BITS, &bitps)) {
                    ifBLIN_QW0("audio bitps %d", c);
                    EXOUT(EX_IOERR);
                }
                if  (0 > ioctl(env->aufd[c], SOUND_PCM_READ_RATE, &rate)) {
                    ifBLIN_QW0("audio rate %d", c);
                    EXOUT(EX_IOERR);
                }
                if  (0 > ioctl(env->aufd[c], SNDCTL_DSP_GETBLKSIZE, &bsize)) {
                    ifBLIN_QW0("audio bsize filt %d", c);
                    EXOUT(EX_IOERR);
                }
                fprintf(fo, " %d x %db x %dHz bsize=%dBytes", chan, bitps, rate, bsize);
            }
            fprintf(fo, "\n");
            if  (full & 4) {
                for (m = 0; m < SOUND_MIXER_NRDEVICES; ++m) {
                    if  ((r[0] | r[1] | r[2] | r[3]) & (1 << m)) {
                        int vol;

                        fprintf(fo, "    |   ");
                        for (i = 0; i < RRR; ++i) {
                            if  (r[i] & (1 << m)) {
                                fprintf(fo, "       %2d", m);
                            } else {
                                fprintf(fo, "       --");
                        }   }
                        if  (r[0] & (1 << m)) {
                            if  (0 > ioctl(env->aufd[c], MIXER_READ(m), &vol)) {
                                ifBLIN_QW0("audio vol %d %d", c, m);
                                EXOUT(EX_IOERR);
                            }
                            fprintf(fo, " %3u", vol & 0x0000FF);
                            vol = (vol >> 8) & 0x0000FF;
                            fprintf(fo, ":%u", vol);
                            if  (vol < 100) {
                                fprintf(fo, " ");
                            }
                            if  (vol < 10) {
                                fprintf(fo, " ");
                            }
                        } else {
                            fprintf(fo, "        ");
                        }
                        fprintf(fo, " %s %s\n", aulabels[m], aunames[m]);
    }   }   }   }   }
    bzero(&sp, sizeof(sp));
    if  (0 > ioctl(env->audev, AIOGFMT, &sp)) {
        ifBLIN_QW0("audio params get");
        EXOUT(EX_IOERR);
    }
    fprintf(fo, "    play=%luHz rec=%luHz ", sp.play_rate, sp.rec_rate);
    aufrm(fo, sp.play_format);
    fprintf(fo, " ");
    aufrm(fo, sp.rec_format);
    bzero(&blocksz, sizeof(blocksz));
    if  (0 > ioctl(env->audev, AIOGSIZE, &blocksz)) {
        ifBLIN_QW0("audio block size get");
        EXOUT(EX_IOERR);
    }
    fprintf(fo, " play=%dBytes rec=%dBytes\n", blocksz.play_size, blocksz.rec_size);
/*****************************************************************************************************
 *  sp.play_rate = sp.rec_rate = rate;
 *  sp.play_format = sp.rec_format = bitps;
 *  if  (0 > ioctl(env->audev, AIOSFMT, &sp)) {
 *      ifBLIN_QW0("audio params set");
 *      EXOUT(EX_IOERR);
 *  }
 *  ifBLIN_QX2("%lu %lu %lu %lu\n", sp.play_rate, sp.rec_rate, sp.play_format, sp.rec_format);
 *****************************************************************************************************/
    bzero(&blocksz, sizeof(blocksz));
    if  (0 > ioctl(env->audev, AIONWRITE, &blocksz)) {
        ifBLIN_QW1("audio non-blocking size");
    } else {
        fprintf(fo, "    ready play=%dBytes rec=%dBytes", blocksz.play_size, blocksz.rec_size);
    }
    syp.chan = AIOSYNC_CAPTURE | AIOSYNC_SELECT | AIOSYNC_FREE;
    syp.chan |= AIOSYNC_PLAY | AIOSYNC_SIGNAL | AIOSYNC_READY;
    syp.pos = 0;
    if  (0 > ioctl(env->audev, AIOSYNC, &syp)) {
        ifBLIN_QW0("audio sync params");
        EXOUT(EX_IOERR);
    }
    fprintf(fo, "    sync %ld(%08lX) %ld\n", syp.chan, syp.chan, syp.pos);
    bzero(&mei, sizeof(mei));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GET_RECSRC_NAMES, &mei)) {
        ifBLIN_QW1("GET_RECSRC_NAMES");
    } else {
        fprintf(fo, " RECSRC  ");
        aumixeinfo(fo, &mei);
        c = 0;
        if  (0 > ioctl(env->audev, SNDCTL_DSP_GET_RECSRC, &c)) {
            fprintf(fo, "\n");
            ifBLIN_QW1("GET_RECSRC");
        } else {
            fprintf(fo, " src=%d\n", c);
    }   }
    bzero(&mei, sizeof(mei));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GET_PLAYTGT_NAMES, &mei)) {
        ifBLIN_QW1("GET_PLAYTGT_NAMES");
    } else {
        fprintf(fo, " PLAYTGT ");
        aumixeinfo(fo, &mei);
        c = 0;
        if  (0 > ioctl(env->audev, SNDCTL_DSP_GET_PLAYTGT, &c)) {
            fprintf(fo, "\n");
            ifBLIN_QW1("GET_PLAYTGT");
        } else {
            fprintf(fo, " tgt=%d\n", c);
    }   }
    if  (!(full & 8)) goto out;
    bzero(&bi, sizeof(bi));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETISPACE, &bi)) {
        ifBLIN_QW1("DSP_GETISPACE");
    } else {
        fprintf( fo, " ISPACES Total=%d Avail=%d %dB fragments, %d Bytes free\n"
               , bi.fragstotal, bi.fragments, bi.fragsize, bi.bytes
               );
    }
    bzero(&bi, sizeof(bi));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETOSPACE, &bi)) {
        ifBLIN_QW1("DSP_GETOSPACE");
    } else {
        fprintf( fo, " OSPACES Total=%d Avail=%d %dB fragments, %d Bytes free\n"
               , bi.fragstotal, bi.fragments, bi.fragsize, bi.bytes
               );
    }
    bzero(&ci, sizeof(ci));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETIPTR, &ci)) {
        ifBLIN_QW1("DSP_GETIPTR");
    } else {
        fprintf( fo, " GETIPTR %d bytes, %d blocks, ptr=%d\n"
               , ci.bytes, ci.blocks, ci.ptr
               );
    }
    bzero(&ci, sizeof(ci));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETOPTR, &ci)) {
        ifBLIN_QW1("DSP_GETOPTR");
    } else {
        fprintf( fo, " GETOPTR %d bytes, %d blocks, ptr=%d\n"
               , ci.bytes, ci.blocks, ci.ptr
               );
    }
    bzero(&oct, sizeof(oct));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_CURRENT_IPTR, &oct)) {
        ifBLIN_QW1("DSP_CURRENT_IPTR");
    } else {
        fprintf( fo, " CURIPTR long samples=%lld fifo_samples=%d\n"
               , oct.samples, oct.fifo_samples
               );
    }
    bzero(&oct, sizeof(oct));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_CURRENT_OPTR, &oct)) {
        ifBLIN_QW1("DSP_CURRENT_OPTR");
    } else {
        fprintf( fo, " CUROPTR long samples=%lld fifo_samples=%d\n"
               , oct.samples, oct.fifo_samples
               );
    }
    bzero(&bmd, sizeof(bmd));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_MAPINBUF, &bmd)) {
        ifBLIN_QW1("DSP_MAPINBUF");
    } else {
        fprintf( fo, " MAPIBUF %"BLIN_X" + %d(%08X)\n"
               , BLIN_I(bmd.buffer), bmd.size, bmd.size
               );
    }
    bzero(&bmd, sizeof(bmd));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_MAPOUTBUF, &bmd)) {
        ifBLIN_QW1("DSP_MAPOUTBUF");
    } else {
        fprintf( fo, " MAPOBUF %"BLIN_X" + %d(%08X)\n"
               , BLIN_I(bmd.buffer), bmd.size, bmd.size
               );
    }
    bzero(&ei, sizeof(ei));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETERROR, &ei)) {
        ifBLIN_QW1("DSP_GETERROR");
    } else {
        fprintf( fo, " ERROR   rec  over=%d ptradj=%u err=%d lasterr=%d errparm=%ld\n"
                     "        play under=%d ptradj=%u err=%d lasterr=%d errparm=%ld\n"
               , ei.rec_overruns, ei.rec_ptradjust, ei.rec_errorcount, ei.rec_lasterror, ei.rec_errorparm
               , ei.play_underruns, ei.play_ptradjust, ei.play_errorcount, ei.play_lasterror, ei.play_errorparm
               );
    }
    c = 0;
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETFMTS, &c)) {
        ifBLIN_QW1("DSP_GETFMTS");
    } else {
        fprintf(fo, " GETFMTS ");
        aufrm(fo, c);
        fprintf(fo, "\n");
    }
    c = 0;
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETCAPS, &c)) {
        ifBLIN_QW1("DSP_GETCAPS");
    } else {
        fprintf(fo, " DSPCAPS ");
        aucaps(fo, c);
        fprintf(fo, "\n");
    }
    bzero(&sc, sizeof(sc));
    if  (0 > ioctl(env->audev, AIOGCAP, &sc)) {
        ifBLIN_QW1("AIOGCAP");
    } else {
        fprintf( fo, " AIOCAPS %lu<=rate<=%lu DMAbuf=%lu mix%08lX L%u R%u\n         "
               , sc.rate_min, sc.rate_max, sc.bufsize, sc.mixers, sc.left, sc.right
               );
        incaps(fo, sc.inputs);
        fprintf(fo, "\n         ");
        aufrm(fo, sc.formats);
        fprintf(fo, "\n");
    }
    k = 0;
    bzero(&ull, sizeof(ull));
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GET_CHNORDER, &ull)) {
        ifBLIN_QW1("GET_CHNORDER");
    } else {
        fprintf( fo, " CHORDER %llX", ull);
        ++k;
    }
    c = 0;
    if  (0 > ioctl(env->audev, SNDCTL_DSP_GETODELAY, &c)) {
        ifBLIN_QW1("DSP_GETODELAY");
    } else {
        fprintf( fo, "%s ODELAY  %d", k++ ? "    " : "", c);
    }
out:
    if  (0 > fflush(fo)) {
        ifBLIN_QW0("fflush");
    }
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_ana(auwith_env *env, FILE *fo) {                                                          /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int c, i, ex = EX_OK;
    oss_mixer_value omv;
    oss_audioinfo oai;
    oss_mixerinfo omi;
    oss_card_info oci;
    oss_midi_info omd;
    oss_sysinfo osi;
    oss_mixext mxt;
    mixer_info mi;
    u_int32_t m;

    c = 0;
    if  (0 > ioctl(env->audev, OSS_GETVERSION, &c)) {
        ifBLIN_QW1("VERSION");
    } else {
        fprintf(fo, " VERSION %d\n", c);
    }
    bzero(&osi, sizeof(osi));
    if  (0 > ioctl(env->audev, SNDCTL_SYSINFO, &osi)) {
        ifBLIN_QW1("SYSINFO");
    } else {
        fprintf( fo
               , " SYSINFO  prod='%s' ver='%s'(%d) opt='%s' Ncard=%d Nsynth=%d Ntim=%d Nmix=%d\n"
                 "          Nau=%d<%08X %08X %08X %08X %08X %08X %08X %08X>\n"
                 "          Nmidi=%d<%08X %08X %08X %08X %08X %08X %08X %08X>\n"
               , osi.product
               , osi.version
               , osi.versionnum
               , osi.options
               , osi.numcards
               , osi.numsynths
               , osi.numtimers
               , osi.nummixers
               , osi.numaudios
               , osi.openedaudio[0], osi.openedaudio[1], osi.openedaudio[2], osi.openedaudio[3]
               , osi.openedaudio[4], osi.openedaudio[5], osi.openedaudio[6], osi.openedaudio[7]
               , osi.nummidis
               , osi.openedmidi[0], osi.openedmidi[1], osi.openedmidi[2], osi.openedmidi[3]
               , osi.openedmidi[4], osi.openedmidi[5], osi.openedmidi[6], osi.openedmidi[7]
               );
    }
    bzero(&oci, sizeof(oci));
    for (i = 0; i < osi.numcards; ++i) {
        oci.card = i;
        if  (0 > ioctl(env->audev, SNDCTL_CARDINFO, &oci)) {
            ifBLIN_QW1("CARDINFO %d", i);
        } else {
            fprintf( fo
                   , " CARDINFO %d\n          shortname='%s' longname='%s' flags=%08X\n"
                   , oci.card
                   , oci.shortname
                   , oci.longname
                   , oci.flags
                   );
    }   }
    bzero(&omi, sizeof(omi));
    for (i = 0; i < osi.nummixers; ++i) {
        omi.dev = i;
        if  (0 > ioctl(env->aumix, SNDCTL_MIXERINFO, &omi)) {
            ifBLIN_QW1("MIXERINFO %d", i);
        } else {
            fprintf( fo
                   , " MIXINFO  %d\n"
                     "          %s id=%s caps=%d modify_counter=%d card_number=%d port_number=%d\n"
                     "          handle=%s magic=%d enabled=%d flags=%d nrext=%d priority=%d\n"
                   , omi.dev
                   , omi.name
                   , omi.id
                   , omi.caps
                   , omi.modify_counter
                   , omi.card_number
                   , omi.port_number
                   , omi.handle
                   , omi.magic
                   , omi.enabled
                   , omi.flags
                   , omi.nrext
                   , omi.priority
                   );
    }   }
    bzero(&oai, sizeof(oai));
    for (i = 0; i < osi.numaudios; ++i) {
        oai.dev = i;
        if  (0 > ioctl(env->audev, SNDCTL_AUDIOINFO, &oai)) {
            ifBLIN_QW1("AUDIOINFO %d", i);
        } else {
            fprintf( fo
                   , " AUDINFO  %d\n          %s devnode=%s\n          "
                   , oai.dev
                   , oai.name
                   , oai.devnode
                   );
            aucaps(fo, oai.caps);
            fprintf(fo, " i");
            aufrm(fo, oai.iformats);
            fprintf(fo, " o");
            aufrm(fo, oai.oformats);
            fprintf(fo, "\n");
            fprintf( fo
                   , "          busy=%d pid=%d magic=%d\n"
                     "          cmd=%s card_number=%d port_number=%d mixer_dev=%d real_device=%d\n"
                     "          enabled=%d %d<rate<%d %d<channels<%d\n"
                     "          binding=%d rate_source=%d handle=%s latency=%d\n"
                     "          song_name=%s label=%s\n"
                     "          [%u]"
                   , oai.busy
                   , oai.pid
                   , oai.magic
                   , oai.cmd
                   , oai.card_number
                   , oai.port_number
                   , oai.mixer_dev
                   , oai.real_device
                   , oai.enabled
                   , oai.min_rate
                   , oai.max_rate
                   , oai.min_channels
                   , oai.max_channels
                   , oai.binding
                   , oai.rate_source
                   , oai.handle
                   , oai.latency
                   , oai.song_name
                   , oai.label
                   , oai.nrates
                   );
            for (m = 0; m < oai.nrates; ++m) {
                fprintf(fo, "%c%u", m ? ',' : '=', oai.rates[m]);
            }
            fprintf(fo, "\n");
    }   }
    bzero(&omd, sizeof(omd));
    for (i = 0; i < osi.nummidis; ++i) {
        omd.dev = i;
        if  (0 > ioctl(env->audev, SNDCTL_MIDIINFO, &omd)) {
            ifBLIN_QW1("MIDIINFO");
        } else {
            fprintf( fo, " MIDINFO dev=%d name=%s busy=%d pid=%d"
                   , omd.dev, omd.name, omd.busy, omd.pid
                   );
            for (m = 0; m < 64; ++m) {
                if  (!(m & 0x0F)) fprintf(fo, "\n        ");
                fprintf(fo, " %02X", omd.cmd[m]);
            }
            fprintf(fo, "\n         ");
            midicaps(fo, omd.caps);
            fprintf(fo, "\n         magic=%d card_N=%d port_N=%d enabled=%d flags=%d handle=%s\n"
                        "         latency=%dus song=%s label=%s\n"
                   , omd.magic
                   , omd.card_number
                   , omd.port_number
                   , omd.enabled
                   , omd.flags
                   , omd.handle
                   , omd.latency
                   , omd.song_name
                   , omd.label
                   );
    }   }
    bzero(&mxt, sizeof(mxt));
    if  (0 > ioctl(env->aumix, SNDCTL_MIX_EXTINFO, &mxt)) {
        ifBLIN_QW1("MIX_EXTINFO");
    } else {
        fprintf( fo
               , " MIX_EXT dev=%d ctrl=%d type=%d minvalue=%d maxvalue=%d %08X id=%s\n"
                 "         parent=%d timestamp=%d data=%s\n"
                 "         %02X %02X %02X %02X  %02X %02X %02X %02X"
                        "  %02X %02X %02X %02X  %02X %02X %02X %02X\n"
                 "         %02X %02X %02X %02X  %02X %02X %02X %02X"
                        "  %02X %02X %02X %02X  %02X %02X %02X %02X\n"
                 "         control_no=%d extname=%s update_counter=%d\n"
               , mxt.dev
               , mxt.ctrl
               , mxt.type
               //, mxt.
               , mxt.minvalue
               , mxt.maxvalue
               , mxt.flags
               //, mxt.
               , mxt.id
               , mxt.parent
               , mxt.timestamp
               , mxt.data
               , mxt.enum_present[ 0], mxt.enum_present[ 1], mxt.enum_present[ 2], mxt.enum_present[ 3]
               , mxt.enum_present[ 4], mxt.enum_present[ 5], mxt.enum_present[ 6], mxt.enum_present[ 7]
               , mxt.enum_present[ 8], mxt.enum_present[ 9], mxt.enum_present[10], mxt.enum_present[11]
               , mxt.enum_present[12], mxt.enum_present[13], mxt.enum_present[14], mxt.enum_present[15]
               , mxt.enum_present[16], mxt.enum_present[17], mxt.enum_present[18], mxt.enum_present[19]
               , mxt.enum_present[20], mxt.enum_present[21], mxt.enum_present[22], mxt.enum_present[23]
               , mxt.enum_present[24], mxt.enum_present[25], mxt.enum_present[26], mxt.enum_present[27]
               , mxt.enum_present[28], mxt.enum_present[29], mxt.enum_present[30], mxt.enum_present[31]
               , mxt.control_no
               , mxt.extname
               , mxt.update_counter
               );
    }
    bzero(&omv, sizeof(omv));
    if  (0 > ioctl(env->aumix, SNDCTL_MIX_READ, &omv)) {
        ifBLIN_QW1("MIX_READ");
    } else {
        fprintf( fo, " MIXREAD dev=%d ctrl=%d value=%d flags=%08X timestamp=%d\n"
               , omv.dev
               , omv.ctrl
               , omv.value
               , omv.flags
               , omv.timestamp
               );
    }
    bzero(&mi, sizeof(mi));
    if  (0 > ioctl(env->aumix, SOUND_MIXER_INFO, &mi)) {
        ifBLIN_QW1("MIXER_INFO");
    } else {
        fprintf( fo, " MIXER   id=%s~ name=%s~ mod=%d\n"
               , mi.id, mi.name, mi.modify_counter
               );
    }
/*  SNDCTL_DSP_GETPLAYVOL, &int */
    if  (0 > fflush(fo)) {
        ifBLIN_QW0("fflush");
    }
    return(ex);
}
