/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: util.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

static int bufcnt = 0;

static const char *
sync[] = {"SYNCNO", "SYLAZY", "SPATIN", "STRICT"};

static const char *caps[] =
{ "unkn"  ,  "unkn"  ,  "SKEW"  , "unkn"
, "INP"   ,  "OUT"   ,  "unkn"  , "unkn"
, "unkn"  ,  "unkn"  ,  "unkn"  , "unkn"
, "ALAW"  ,  "uLAW"  ,  "unkn"  , "unkn"

, "HARDTI",  "LOOK"  ,  "DECONF", "DUMPOP"
, "unkn"  ,  "unkn"  ,  "unkn"  , "USED"
};

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_capand(BLIN_flag flags, FILE *fo, int kind) {                                             /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int32_t i, m, b;
    int k;

    b = 0;
    fprintf(fo, "%08X", flags);
    if  (!!kind) fprintf(fo, "%c%s", b++ ? ',' : '<', sync[flags & AUWITH_CAP_SMASK]);
    if  (!!(flags & AUWITH_CAP_SWAP))
        fprintf(fo, "%cswap=%02X", b++ ? ',' : '<', (flags & AUWITH_CAP_SWAP) >> AUWITH_CAP_SWAPS);
    flags &= ~AUWITH_CAP_SWAP;
    for (m = 0x04, i = 2; i < 24; ++i, m <<= 1) {
        if  (flags & m) fprintf(fo, "%c%s", b++ ? ',' : '<', caps[i]);
    }
    for (m = BLIN_BIT7, k = 7; k >= 0; --k, m >>= 1) if (flags & m) break;
    if  (k >= 0) fprintf(fo, "%cBIT%d", b++ ? ',' : '<', k);
    fprintf(fo, ">");
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_caproc(auwith_plug *plug, FILE *fo) {                                                     /****
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int32_t b;

    b = 0;
    if  (!!plug->parsin) fprintf(fo, "%cparsin", b++ ? ',' : '<');
    if  (!!plug->deconf) fprintf(fo, "%cdeconf", b++ ? ',' : '<');
    if  (!!plug->activ) fprintf(fo, "%cactiv", b++ ? ',' : '<');
    if  (!!plug->passiv) fprintf(fo, "%cpassiv", b++ ? ',' : '<');
    if  (!!plug->reed) fprintf(fo, "%creed", b++ ? ',' : '<');
    if  (!!plug->rite) fprintf(fo, "%crite", b++ ? ',' : '<');
    fprintf(fo, ">");
}

auwith_time
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_future(auwith_time time) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    struct timespec tn;
    auwith_time t;

    if  (0 > clock_gettime(CLOCK_UPTIME_PRECISE, &tn)) {
        ifBLIN_QW0("clock_gettime");
        t = 0;
        goto out;
    }
    t = (u_int64_t)tn.tv_sec * 1000000000ULL + tn.tv_nsec + time;
out:
    return(t);
}

auwith_time
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_bufuture(auwith_ent *ent, int64_t v) {                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int64_t samplewidth = (ent->flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW)) ? 1 : (ent->bits + 7) / 8;

    return(auwith_future(v * 1000000000 / (samplewidth * ent->rate * ent->nchan)));
}

auwith_time
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_fifuture(auwith_ent *ent) {                                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
    return(auwith_future(1000000000LL * ent->block / ent->rate));
}

auwith_buf *
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_creabuf(size_t size) {                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    auwith_buf *buf = NULL;

    if  (!(buf = calloc(1, size + sizeof(auwith_buf)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
    } else {
        buf->refcnt = 1;
        buf->size = size;
        ifBLIN_QX7("buf %d", ++bufcnt);
    }
    return(buf);
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_relebuf(auwith_buf *buf) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (0 >= --buf->refcnt) {
        free(buf);
        ifBLIN_QX7("buf %d", --bufcnt);
}   }

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_takebuf(auwith_buf *buf) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (!!buf) {
        ++buf->refcnt;
        ifBLIN_QX7("buf %d", ++bufcnt);
}   }
